/*
   Class: dForm
   Author:  Daniel Howard (sproket@total.net)
   History: 
      1.0   November 21, 1998
         Initial Release

      1.1   March 27, 1999
         Additions:
         * Added an OnGot/LostFocus to attach this to _app.ActiveForm.
           useful for debugging from the command window.
         * Added an auto cascade if there is a CallingForm object
           defined.
         * Added the Freeze and UnFreeze API calls.

      1.2 April 28, 1999
         * Removed onLostFocus

   Inherits: FORM

   Properties:

      ParentClassName -
               Assigned to conform do the dControls. All objects 
               subclassed from this form (or any dControl) have it.
               It's simply to allow you to know what kind of an object
               this is no matter how far down you subclass.
      
   Methods:

      Init  -  Abstact method. To be overridden. Used to execute any 
               initialization code "before" the form opens. 
               Thanks to Peter Rorlick for the concept.

      Iterate- Used to execute a code block or function pointer 
               against all controls on the form. 

               eg.
               cb = {|o|; o.FontBold = true} (Don't forget the semicolon)
               form.Iterate(cb)
               OR
               fp = Class::DoWhatever
               form.Iterate(fp)

      Repaint - Calls Win API to repaint the form.

   Overrides:

      ReadModal - Calls custom OpenMe method for handling init methods.

      Open      - See ReadModal.

      OnGotFocus- Assigns form to _app.ActiveForm

   Notes:
   DO NOT DESIGN THIS FORM! Certain lines will be killed by the
   designer. There is no need to design this form anyway because 
   there is no visual constructor code.


This software is donated to the public domain. You may modify 
this code in anyway you wish however the author reserves the 
right to whatever official version is published. If you have 
further ideas/bug repors/additions then contact the author and 
your name will be added to the contributors if your idea is 
implemented.

The author accepts no responsibility for the use or consequences 
of the use of this software.
*/

class dForm of FORM custom
Protect OpenMe

   with (this)
      scaleFontSize = 8
   endwith


   // Needs to be here -- careful of designer!
   this.ParentClassName = "FORM"

   Function Open
      Class::OpenMe()
   return ( Super::Open() )

   Function ReadModal
      Class::OpenMe()
      this.MDI := false
		Super::ReadModal()
   return


   // The OpenMe method evaluates possible INIT methods that may
   // exist within the object hierarchy. 
   /* Calling Order:
   + Form
   \->Init()            First
      +  Container
      \->Init()         Next
         +  Control
         \->Init()      Last
   */
   Function OpenMe
      if type("this.Init") # "U"
         this.Init()
      endif   

      // This is an example of using the Iterate method of the form.
      // I will look at every control on the form for an INIT method 
      // and execute it if one is found.               
      Class::Iterate({|o|; iif(type("o.Init") # "U", o.Init(),false) })

      if type("this.CallingForm") = "O"
         this.AutoCenter := false
         this.top := this.CallingForm.Top + 1
         this.left := this.CallingForm.Left + 2
      endif
   return


   // Simple enough. I set a property to _app. This is very useful
   // for debugging from the command window.
   Function OnGotFocus
      _app.ActiveForm = this
   return

   // This method iterates though the form executing a codeblock
   // for each object recursively.
   // Be careful of syntax or other errors in your codeblocks
   // since Visual dBASE will point here if there is an error,
   // NOT in the codeblock itself. This could cause some headaches
   // if you have a lot of Init methods or other itineration code.
   // I left in a ?cb statement to help you debug.

   // Normally takes 1 parameter but you can also pass the starting 
   // object. The default is FORM.

   // This method depends on whether there is a first property defined 
   // to know if an object is a container or notebook.
   Function Iterate
   para cb,o

      if type("o") # "O"
         o = form
      endif

      private oControl
      oControl = o.First
      if oControl = NULL
         return
      endif
      do
         if type("cb") = "FP" or type("cb") = "CB"
            cb(oControl)
         endif
         if type("oControl.first") = "O"
            Class::Iterate(cb,oControl)      
         endif                   
         oControl = oControl.Before
      until oControl = o.First
   return

   Procedure Repaint
      if type("GetDesktopWindow") # "FP"
         extern CHANDLE GetDesktopWindow() User32
      endif

      if type("UpdateWindow") # "FP"
         extern CLOGICAL UpdateWindow(CHANDLE) User32
      endif

      UpdateWindow(this.hWND)
   return

   Function Freeze
      if Type("LockWindowUpdate") # "FP"
          extern CLOGICAL LockWindowUpdate(CHANDLE) User32
      endif

      If Type("GetD*fesktopWindow") # "FP"
          extern  CHANDLE   GetDesktopWindow ( CVOID )  USER32
      Endif

      LockWindowUpdate(this.hWnd)  && Freeze
   return
  
   Function UnFreeze
      If Type("LockWindowUpdate") # "FP"
         extern CLOGICAL LockWindowUpdate(CHANDLE) User32
      EndIf
      LockWindowUpdate(0)  && UnFreeze
   return

endclass
