/*
Class:   TreeDropper
Author:  Daniel Howard (sproket@total.net)
         Written for Visual dBASE Version 7.01

History: 
   1.0 Oct 31, 1998 
      ( Beta )

   1.1 Nov 07, 1998
   -  Cleaned up & tested.
   -  The treeview is now created at the form level to ensure that it
      will appear on top of all other controls no matter what the 
      zOrder of the container itself is.
   -  Added an onChange event handler.
   -  Fixed up some wierd behaviour with right clicking.
   -  Added a TreeDropper custom property to be used instead of parent 
      since the TreeView's parent is the form. 

   1.2 Dec 12, 1998
   - Got rid of the TreeDropper custom property as it is now called
     setter and is defined in dContainer.

   1.3 Apr 23, 1999
   - Removed setter/getter technology as it crashes the form
     designer sometimes.
   - Added the Win API SetParent function to handle the z-order
     issue.
	- Changed Seperator to SEPARATOR. Duh Spelling...

Purpose:
   Use this object like a combobox except the dropdown is not a simple
   list but a treeview object. This is similar to the kind of object 
   used in the Vdb Inspector.

   This class requires:
   dForm.cfm      - Base form class 
   down.bmp       - Down Arrow bitmap 

   Please see dForm.cfm for the specific behaviours required in that 
   classes. You can use this code in you own base  clases. (you DO 
   have your own base classes right?)


   Properties:

      DropDownHeight- Specifies the dropdown height of the TreeDropper 
                      in CHAR metrics.

      ShowParents -   If this property is set to true then the value 
                      of TreeDropper with include the hierarchy.

      Separator   -  This is the separator between parents and children 
                     shown in the entryfield. It defaults to "."

      Selected    -  A refrence to the selected treeitem. 

      IsOpen      -  A flag to indicate whether the treeview is open or not.


Form Properties
   ActiveTreeDropper - A reference to the currently opened TreeDropper

Other properties:

   SavedInfo         -  A simple collection object to hold overridden
                        values to be restored later.
Notes:
   - Currently this only works with metric CHARS.


This software is donated to the public domain. You may modify 
this code in anyway you wish however the author reserves the 
right to whatever official version is published. If you have 
further ideas/bug reports/additions then contact the author and 
your name will be added to the contributors if your idea is 
implemented.

The author accepts no responsibility for the use or consequences 
of the use of this software.
*/

class TreeDropper(parentObj) of CONTAINER(parentObj) custom

   with (this)
      left = 0.8333
      top = 4.4
      width = 49
      height = 1.8
      metric = 0        // Chars
   endwith


   this.EntryField = new EntryField(this)
   with (this.EntryField)
      onLeftMouseDown = class::ShowOrHideDropDown
      height = 1
      left = 0.3333
      top = 0.05
      width = 47.1667
      value = ""
		When = {; return false}
		MousePointer = 1
   endwith

   this.TREEVIEW = new TD_TREEVIEW(this)
   with (this.TREEVIEW)
      onLeftMouseDown = Class::TREEVIEW_onLeftMouseDown
      onRightMouseDown = Class::TREEVIEW_onRightMouseDown
      height= 0.25
      left = 0.3333
      top = 1.05
      width = 47.1667
      hasButtons = false
      hasLines = false
   endwith

   this.IsOpen = false
   this.ShowParents = false
   this.Separator = "."
   this.DropDownHeight = 7

   this.Selected = NULL

   // Storage space for temporary property values.
   this.SavedInfo = New Object()

   form.ActiveTreeDropper = NULL

   Function Init

      // Resize the entryfield to match the width of the container
      this.EntryField.width = this.width - 1.83 

      // Create a pushbutton at the right edge of the entryfield.
      this.DropBUTTON  = new PUSHBUTTON(this)
      with (this.DropButton)
         onClick = class::ShowOrHideDropDown
         height = this.EntryField.Height - 0.15
         left = this.EntryField.width - 3
         top = 0.15
         width = 3
         text = ""
         value = false
  			speedBar = true
  			upBitmap = "filename Down.bmp"
      endwith

      // Prototype the API 
      if type("SetParent") = "U"
         extern CHANDLE SetParent(CHANDLE, CHANDLE) user32         
      endif

      this.BorderStyle = 3 // None

      // Let's set some properties for the treeview that we know 
      // aren't needed.
      with (this.TreeView)
         visible = false
         allowEditLabels = false
         allowEditTree = false
         disablePopUp = true
      endwith 
   return


   // If we're here than a selection has been made!
   function TREEVIEW_OnLeftMouseDown(flags, col, row)
      // Assign the selection to the container
      this.Parent.Selected := this.Selected

      // Resize the parent container.
      this.Parent.Height = this.Parent.SavedInfo.Height

      // Kill the stupid hand cursor.
      this.mousePointer := 0

      // Assign to the entryfield
      if type("this.Selected") = "O"
         local cText,o
         if this.parent.ShowParents
            cText = this.Parent.Selected.Text
            o = this.parent.Selected
            do while o.Parent.ClassName == "TREEITEM"
               o = o.Parent
               cText = o.Text + this.Parent.Separator + cText
            enddo
         else
            cText = this.selected.Text
         endif

         this.Parent.EntryField.Value := cText
      endif

      Class::hide()

      // Check for onChange
      local lSelectionChanged
      lSelectionChanged = false

      if type("this.parent.savedInfo.selected") = "O"
         if this.parent.savedInfo.selected # this.Selected
            lSelectionChanged := true
         endif            
      else
         lSelectionChanged := true
      endif

      if lSelectionChanged and type("this.parent.onChange") # "U"
         this.parent.onChange()
      endif
   return


   // This is fired by the entryfield or the pushbutton.
   function ShowOrHideDropDown
      if this.Parent.IsOpen // Hide it 
         Class::Hide()
      else                  // Show it
         Class::Show()
      endif
   return

   // Can be called outside as well to manually hide the dropdown.
   Function Hide
      if form.ActiveTreeDropper # NULL
         form.ActiveTreeDropper.TreeView.Visible := false
         form.ActiveTreeDropper.Height = form.ActiveTreeDropper.SavedInfo.Height
         form.ActiveTreeDropper.IsOpen := false
         form.ActiveTreeDropper.TreeView.MousePointer := 0
         form.onLeftMouseDown = form.ActiveTreeDropper.SavedInfo.OnLeftMouseDown
      endif
   return         

   // Can be called outside as well to manually show the dropdown.
   Function Show

      // Check if the ActiveTreeDropper is different to this one. If it is
      // then we hide the old one.
      if form.ActiveTreeDropper # NULL and form.ActiveTreeDropper # this.parent
         form.ActiveTreeDropper.Hide()
      endif

      // Tell the form what the active TreeDropper is
      form.ActiveTreeDropper = this.Parent

      // Call the API.
      SetParent(form.ActiveTreeDropper.hWnd,form.hWnd)
              
      // Resize the treeview according to the dropdownheight property.
      this.Parent.SavedInfo.Height = this.Parent.Height
      this.Parent.Height = this.Parent.DropDownHeight
      this.Parent.TreeView.Width = this.Parent.Width - 1.83
      this.Parent.TreeView.Height = this.Parent.Height - 1

      // This block makes sure that all children are expanded.
      this.Parent.TreeView.Iterate({|o|; o.Expanded := true})

      // Save the oldSelected for test in OnChange
      // Not really needed since the treedropper propety handles this.
      this.Parent.SavedInfo.Selected = this.Parent.Selected

      // This block of code is to force the treeview to jump to 
      // the appropriate place in the list. Although I force a 
      // select to the 1st item in the treeview I don't assign it
      // to the TreeDropper itself. This is to ensure that if there
      // is no current selection, the treeview will display properly.
      if this.Parent.Selected = NULL
         if this.Parent.TreeView.FirstChild # NULL
            this.Parent.TreeView.FirstChild.Select()
         endif
      else
         this.Parent.Selected.Select()
      endif

      // Save the onLeftMouseClick of the form itself and attach our
      // hide method to it.
      this.Parent.SavedInfo.OnLeftMouseDown = form.onLeftMouseDown
      form.onLeftMouseDown = Class::Hide


      // Open it up
      this.Parent.TreeView.Visible = true
      this.Parent.TreeView.SetFocus()
      this.Parent.IsOpen = true
   return

   // This is assigned because of some anomolous behavours of the 
   // treeview. Normally, Right clicking selects a treeitem but 
   // doesn't assign the selected property. Also Right and Left 
   // mousedown events seem to be equivilant unless an event handler
   // is defined.  So I bloody well define one here!
   function TREEVIEW_OnRightMouseDown(flags, col, row)
   return

endclass



//Class TD_EntryField(parentObj) of EntryField(parentObj)
//protect when,MousePointer
//   this.When = {; return false}
//   this.MousePointer = 1
//endclass
//
//Class TD_PushButton(parentObj) of PushButton(parentObj)
//protect speedbar,upBitMap
//  this.speedBar = true
//  this.upBitmap = "filename Down.bmp"
//endclass
//


/*   
   Methods:

      Iterate  -  Loops though all treeitems in the treeview and 
                  executes a code block or function pointer against 
                  each one.

                  Note: This method doesn't work if called in an Init
                  method of a custom treeview. I don't know if this is 
                  a wad or a bug but the firstChild property doesn't 
                  get set until after the treeview has been opened. 
                  The work around is to call it in an OnOpen event of 
                  the treeview.

      Fill     -  Takes an array as a parameter and fills the treeview 
                  with the array values. See the TreeDropper sample 
                  form to see how this can work. You can specify 
                  elements of the array as TreeItemInfo objects if you
                  need specific settings for the treeitems to be 
                  applied like image or bold. This method also creates 
                  a "Datasource" property which will be an array of 
                  the same structure that was passed containg pointers
                  to the actual tree items.

                  Note: ALWAYS PASS THE ARRAY BY VALUE!!!
                  a = {"one","two","three"}
                  TreeView.Fill( (a) )   // Extra BRACKETS!
                  Otherwise you're bound to get unpredictable results!
                  (At least I did!)

   Notes:

      There are a few issues with my TreeView. 

      1) If you use the interactive treeview make sure you have no
         treeitems defined in the form designer. There seems to be
         a bug with ReleaseAllChildren() which prevents the treeview
         from completely releasing them. It works ok as long as all
         treeitems are created at run-time. 

*/
Class TD_TreeView(parentObj) of TreeView(parentObj)

   // Use this custom method of the treeview to iterate through all 
   // items and perform a code block on each.
   Function Iterate
   para cb, o

      if argcount() = 1
         o = this.firstchild
      endif

      private oControl
      oControl = o
      do while oControl # NULL
         cb(oControl)
         if oControl.NoOfChildren > 0
            local oChild
            oChild = oControl.FirstChild
            this.Iterate(cb,oChild)      
         endif      
         oControl = oControl.NextSibling
      enddo
   return 

   // This will take an array as a parameter.
   // Each element can be either a string or another array.
   // This method will fill the treeview according to the array 
   // structure.
   // ReleaseAllChildren() should be called first.

   /* Array Sample
         Test1
         Test2
         { "item1","item2" }
         Test3
         { "item4",{"item5","item6"} }

      Shold look like
      Test1
      Test2
         Item1
         Item2
      Test3
         Item4
            Item5
            Item6                              
   */
   Function Fill
   para a, oParent, aDataSource

      if argcount() = 1 // Top Level
         oParent = this
         this.DataSource = New Array()
         aDataSource = this.DataSource
      endif

      local j, t, cName

      private xItem

      t = oParent
      local n
      n = 1
      for j = 1 to a.Size
         xItem = a[j]
         do case
         case type("xItem") == "A"
            local a2
            a2 = new array(a[j].size)
            acopy(a[j],a2)
            aDataSource.Add( New Array() )
            this.Fill( (a2), t , aDataSource[aDataSource.Size] )
            n = j

         case type("xItem") == "C"
            cName = "TreeItem" + n
            t = New TreeItem(oParent,cName)
            t.Text = a[j]
            aDataSource.Add(t)
            n++

         case type("xItem") == "O"
            cName = "TreeItem" + j
            t = New TreeItem(oParent,cName)
            // This is a TreeItemInfo Object which holds 
            // information about a treeitem. 
            if a[j].ClassName # "TREEITEMINFO"
               t.Text := "Invalid Tree Item Passed!"
            else
               if not empty(a[j].Name)
                  t.Name := a[j].Name
               endif
               t.Bold := a[j].Bold
               t.Checked := a[j].Checked
               t.Image := a[j].Image
               t.SelectedImage := a[j].SelectedImage
               t.Text = a[j].Text
               t.Expanded := a[j].Expanded
               aDataSource.Add(t)               
            endif
            n++
         endcase
      next j
   return

/*
   Routine to let you add an item to a treeview programically
   without having to deal with naming the damn things.

   I use this logic to add an item:
      if you had 
         treeitem1
         treeitem3
      then you would get
         treeitem1
         treeitem3
         treeitem2

   I load the numerical part of the treeitems into an array and 
   then use a.Scan() to see if a treeitem exists.  I double check
   using the type() function in case you create treeitems with non
   standard names.

   This is consitant with the way the treeview works normally.

   The 2nd parameter is the parent to add to. If none is 
   specified then the parent is the treeview itself.
   
   This returns a reference to the newly added treeitem.
*/
   Function AddItem
   para cText,oParent

      if argcount() = 1
         oParent = this
      endif

      private o
      local a
      a = New Array()
      o = oParent.FirstChild
      do while o # NULL
         // Extract the number from the treeitem's name and 
         // add it to our array for testing
         a.Add( Val(Stuff(o.Name,1,8,"")) )
         o = o.NextSibling
      enddo

      // This should give us the first tree number that doesn't 
      // already exist.
      private nTreeNo
      nTreeNo = 1
      do while true
         if a.Scan(nTreeNo) = 0
            o = oParent
            if type("o.treeitem"+nTreeNo) = "U"
               exit  // Ok, we got a good one
            endif
         endif
         nTreeNo++
      enddo

      local t,cName
      cName = "TREEITEM"+nTreeNo
      t = New TreeItem(oParent,cName)   
      t.Text := cText
   return ( t )

endclass



/*   
   Class: TreeItemInfo
   Author:  Daniel Howard (sproket@total.net)
   History: 
      1.0   November 23, 1998
            Initial release

   This class is simply a container for treeitem
   information. If you pass an array of these objects
   to the fill method of the dTreeView Class it will 
   use this information to fill the treeitems.
*/
Class TreeItemInfo of dObject
   this.Name = ""
   this.Bold = false
   this.Checked = false
   this.Expanded = false
   this.Image = ""
   this.SelectedImage = ""
   this.Text = ""
endclass

